-- ----------
-- txid.sql
--
--	SQL script for loading the transaction ID compatible datatype 
--
--	Copyright (c) 2003-2004, PostgreSQL Global Development Group
--	Author: Jan Wieck, Afilias USA INC.
--
-- ----------

set client_min_messages = 'warning';

CREATE DOMAIN txid AS bigint CHECK (value > 0);

--
-- A special transaction snapshot data type for faster visibility checks
--
CREATE OR REPLACE FUNCTION txid_snapshot_in(cstring)
	RETURNS txid_snapshot
	AS '$libdir/txid' LANGUAGE C
	IMMUTABLE STRICT;
CREATE OR REPLACE FUNCTION txid_snapshot_out(txid_snapshot)
	RETURNS cstring
	AS '$libdir/txid' LANGUAGE C
	IMMUTABLE STRICT;

--
-- The data type itself
--
CREATE TYPE txid_snapshot (
	INPUT = txid_snapshot_in,
	OUTPUT = txid_snapshot_out,
	INTERNALLENGTH = variable,
	STORAGE = extended,
	ALIGNMENT = double
);

CREATE OR REPLACE FUNCTION get_current_txid()
	RETURNS bigint
	AS '$libdir/txid', 'txid_current' LANGUAGE C
	SECURITY DEFINER;

CREATE OR REPLACE FUNCTION get_current_snapshot()
	RETURNS txid_snapshot
	AS '$libdir/txid', 'txid_current_snapshot' LANGUAGE C
	SECURITY DEFINER;

CREATE OR REPLACE FUNCTION get_snapshot_xmin(txid_snapshot)
	RETURNS bigint
	AS '$libdir/txid', 'txid_snapshot_xmin' LANGUAGE C
	IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION get_snapshot_xmax(txid_snapshot)
	RETURNS bigint
	AS '$libdir/txid', 'txid_snapshot_xmax' LANGUAGE C
	IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION get_snapshot_active(txid_snapshot)
	RETURNS setof bigint
	AS '$libdir/txid', 'txid_snapshot_active' LANGUAGE C
	IMMUTABLE STRICT;


--
-- Special comparision functions used by the remote worker
-- for sync chunk selection
--
CREATE OR REPLACE FUNCTION txid_in_snapshot(bigint, txid_snapshot)
	RETURNS boolean
	AS '$libdir/txid', 'txid_in_snapshot' LANGUAGE C
	IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION txid_not_in_snapshot(bigint, txid_snapshot)
	RETURNS boolean
	AS '$libdir/txid', 'txid_not_in_snapshot' LANGUAGE C
	IMMUTABLE STRICT;

-- ----------
-- txid.sql
--
--	SQL script for loading the transaction ID compatible datatype 
--
--	Copyright (c) 2003-2004, PostgreSQL Global Development Group
--	Author: Jan Wieck, Afilias USA INC.
--
-- ----------

--
-- now the epoch storage
--

CREATE SCHEMA txid;

-- remember txid settings
-- use bigint so we can do arithmetic with it
create table txid.epoch (
	epoch bigint,
	last_value bigint
);

-- make sure there exist exactly one row
insert into txid.epoch values (0, 1);


-- then protect it
create function txid.epoch_guard()
returns trigger as $$
begin
    if TG_OP = 'UPDATE' then
	-- epoch: allow only small increase
	if NEW.epoch > OLD.epoch and NEW.epoch < (OLD.epoch + 3) then
	    return NEW;
	end if;
	-- last_value: allow only increase
	if NEW.epoch = OLD.epoch and NEW.last_value > OLD.last_value then
	    return NEW;
	end if;
    end if;
    raise exception 'bad operation on txid.epoch';
end;
$$ language plpgsql;

-- the trigger
create trigger epoch_guard_trigger
before insert or update or delete on txid.epoch
for each row execute procedure txid.epoch_guard();

